; Card Reader
;
; Written by A M Errington
; Device PIC 12C509 (8 pins)
; Clock speed 1Mhz Tcy=1us
; Resonator 4MHz internal RC
; Reset circuit Internal MCLR
; Watchdog Disabled
;
; Uses Panasonic track 2 card reader or any typical magnetic
; card reader mechanism. Sends data out serially at 1200bd.
;
; This software is Copyright 1998 Andrew M Errington
;
; This file is best viewed with hard tabs set to 4 character
; spacing.
#include "P12c509A.inc"
	TITLE "Card Reader"
	LIST C=120, b=4
	LIST P=12C509A
	__CONFIG   _IntRC_OSC & _WDT_OFF & _CP_OFF & _MCLRE_OFF
	
	ERRORLEVEL -305 ; Suppress "Using default destination of 1 (file)."

; **************************************************************
; General Equates

; PIC12C509A standard registers
INDF EQU 0x00
TMR0 EQU 0x01
PC EQU 0x02
STATUS EQU 0x03
FSR EQU 0x04
OSCCAL EQU 0x05
GPIO EQU 0x06 ; lower 5 bits only
; I/O port bits
GP0 EQU 0x00
GP1 EQU 0x01
GP2 EQU 0x02 ; Shared with T0CKI
GP3 EQU 0x03 ; Always input, shared with MCLR, Vpp
GP4 EQU 0x04 ; Shared with OSC2
GP5 EQU 0x05 ; Shared with OSC1, clkin

; Status register bits

C EQU 0x00 ; Carry flag
DC EQU 0x01 ; Digit carry flag
Z EQU 0x02 ; Zero flag
PD EQU 0x03 ; Power down flag
TO EQU 0x04 ; WDT timeout flag
PA0 EQU 0x05 ; Program page select
GPWUF EQU 0x07 ; GPIO reset bit
; Other control bits
RP0 EQU 0x05 ; Register page select


; Other useful constants
LSB EQU 0x00
MSB EQU 0x07
; Useful PIC macros
; The BTFSS and BTFSC instructions can be confusing to read. It
; is easier to read the code with macros for IFSET and IFCLR
; meaning the opposite of SKPCLR and SKPSET respectively.

#DEFINE IFSET BTFSC
#DEFINE IFCLR BTFSS

#DEFINE IFZ SKPNZ
#DEFINE IFNZ SKPZ
#DEFINE IFC SKPNC
#DEFINE IFNC SKPC

#DEFINE SKPSET BTFSS
#DEFINE SKPCLR BTFSC
; **************************************************************
; Card reader constants

start_code EQU b'01011000' 	; Start Sentinel bit pattern,
							; shifted up into the top 5 bits

end_code EQU b'11111000' 	; End Sentinel bit pattern,
							; shifted up into the top 5 bits
; PIC12C509A RAM location usage
buf_ptr EQU 0x07 ; card data buffer pointer (nibbles)
num_chr EQU 0x08 ; Number of characters read from card
count EQU 0x09 ; General 8 bit counter
flag EQU 0x0A ; Control flags
char_buf EQU 0x0B ; Character buffer, input and serial output
parityLRC EQU 0x0C ; Parity/LRC workspace
temp EQU 0x0D ; Temporary workspace
; EQU 0x0E ; unused
; EQU 0x0F ; unused
lo_mem EQU 0x10 ; Memory buffer start address:
; EQU 0x11 ; Track 2 of the magnetic card contains
; EQU 0x12 ; at most 40 4-bit characters, including
; EQU 0x13 ; the start sentinel, end sentinel and
; EQU 0x14 ; LRC, so 20 bytes are reserved to store
; EQU 0x15 ; all of them. In fact only 37 nibbles
; EQU 0x16 ; are used as the start and end sentinels
; EQU 0x17 ; and the LRC are never stored in this
; EQU 0x18 ; application
; EQU 0x19
; EQU 0x1A
; EQU 0x1B
; EQU 0x1C
; EQU 0x1D
; EQU 0x1E
; EQU 0x1F

; 0x20 to 0x2F are mapped to 0x00 to 0x1F, so the buffer
; continues from 0x30 onwards

; EQU 0x30
; EQU 0x31
; EQU 0x32
hi_mem EQU 0x33 ; Memory buffer end
; EQU 0x34 ; unused
; EQU 0x35 ; unused
; EQU 0x36 ; unused
; EQU 0x37 ; unused
; EQU 0x38 ; unused
; EQU 0x39 ; unused
; EQU 0x3A ; unused
; EQU 0x3B ; unused
; EQU 0x3C ; unused
; EQU 0x3D ; unused
; EQU 0x3E ; unused
; EQU 0x3F ; unused
; Derived constants
; buf_sz is the actual number of nibbles available in the
; buffer. If the buffer continues into Bank 1 care must be
; taken to correct for the discontinuity in address space.
; Here the assembler signals an error if portions of the buffer
; are in the wrong banks.
if lo_mem > 0x1F
	ERROR "Buffer start address (lo_mem) must be in Bank 0"
endif

if hi_mem > 0x1F && hi_mem < 0x30
	ERROR "Buffer end address (hi_mem) must be in upper half of Bank 1"
endif

if hi_mem <= 0x1F
buf_sz EQU ((hi_mem - lo_mem) + 1) * 2
else
buf_sz EQU ((hi_mem - lo_mem) - .15) * 2
endif
; Flag register bit meanings

; bit 7 6 5 4 3 2 1 0 -> found start sentinel
; | | | | | | +-----> found end sentinel
; | | | | | +--------> bad parity
; | | | | +-----------> bad LRC
; | | | +--------------> reached end of buffer
; | | +-----------------> R/^W flag for buffer operations
; | +--------------------> unused
; +-----------------------> unused
found_start EQU 0
found_end EQU 1
bad_parity EQU 2
bad_LRC EQU 3
buf_end EQU 4
read_buf EQU 5

; ParityLRC register bit meanings

; bit 7 6 5 4 3 2 1 0 -> unused
; | | | | | | +-----> unused
; | | | | | +--------> unused
; | | | | +-----------> LRC bit 0
; | | | +--------------> LRC bit 1
; | | +-----------------> LRC bit 2
; | +--------------------> LRC bit 3
; +-----------------------> parity bit

; Note: Later code relies on these bits remaining in this position


; I/O pin declarations

; The card reader connects to +5V and 0V, and has three signal
; lines connected to the following I/O pins:

ser_out EQU GP0 ; serial TxD pin to host
card EQU GP1 ; ^CLD signal (low when card present)
clock EQU GP2 ; ^RCL signal (low when data valid)
signal EQU GP3 ; ^RDT signal from magstripe

; GP4 and GP5 are still available for I/O or for a crystal if
; required.

; Compilation options

; The invert_tx option changes the sense of the ser_out line

invert_tx EQU 0 ; 0 = Idle (logical '1') is 0V


; **************************************************************
; Program code starts here
; **************************************************************

; The PIC12C509 reset vector jumps to top of memory, then the
; program counter rolls to 0x00 after loading the RC osc.
; calibration value into W

	ORG 0x00
	MOVWF OSCCAL
	GOTO start
; **************************************************************
; Subroutines
;
; **************************************************************
send_char
; Call send_char with an ASCII character code in W. This is a
; simple serial output routine which sends the character out
; serially on an output pin at 1200 baud, 8 data bits, no parity,
; 1 stop bit. Assume the PIC oscillator is running at 4MHz.
;
; The baud rate of 1200 baud was chosen as it will work with the
; 12C509 internal RC oscillator generating the timing. Higher
; baud rates require tighter timing tolerance, and will therefore
; require a crystal.
;
; Normally serial communication (RS232) requires a negative
; voltage between -5V to -15V to represent a '1', and a positive
; voltage between +5V and +15V to represent a '0'. Most PC
; serial ports will switch at +/-3V, and in fact will often work
; with 0V and 5V, so it is possible to use a PIC I/O pin, set
; high for a logic '0' and low for a logic '1'. A 1k resistor
; placed in series with the Tx line will limit the current, and
; This is probably acceptable for experimental purposes. For
; robustness, however, it may be desirable to include level shift
; IC, such as the MAX232A from Maxim. The invert_tx compilation
; option can be used to alter the sense of transmitted bits if
; necessary.
;
; At 1200 baud each bit cell is just over 833us in length. At
; power up the serial output line is set to its idle state (logic
; '1'). At the start of transmission it is taken to logic '0'
; for one bit time to generate the start bit. Next, the 8 bits
; of character data are shifted out, LSB first, by rolling them
; down into the carry. The program sets or clears the serial
; line pin according to whether the carry represents a logic '0'
; or '1'. Finally the line is held at logic '1' for at least one
; bit time for the stop bit. The line then rests at this state
; (idle) until it is time to send the next byte
;
; Bit cell timing is done by counting clock cycles: 1 instruction
; is 1us, jumps and skips are 2us.

	MOVWF char_buf 	; Store the character code (in W)
					; to character buffer

	MOVLW .10 		; Set the number of bits (including
	MOVWF count 	; start and stop bits) in count

	CLRC 			; Clear carry because the start bit
					; is a '0'
bit_loop
	IFNC ; serial pin logic '0'
if invert_tx
	BCF GPIO,ser_out
else
	BSF GPIO,ser_out
endif

	IFC ; serial pin logic '1'
if invert_tx
	BSF GPIO,ser_out
else
	BCF GPIO,ser_out
endif

	CALL bit_delay ; Make up the bit time to 833us

	RRF char_buf 	; Roll LSB of char_buf into carry,
					; and the '1' from the bit_delay
					; routine into the MSB

	DECFSZ count ; Loop until all bits have been
	GOTO bit_loop ; shifted out.
	RETLW 0

bit_delay
; The bit length should be 833us in total (=833 Tcy). This
; routine and the bit loop routine take 21 Tcy, leaving 812 Tcy
; to waste. The delay loop is 4 cycles long, so loop for 203
; times. This is done by loading a counter with 255 - 203 = 52
; and incrementing it every time around the loop. When the
; counter reaches 255 it overflows and sets the carry flag. As
; a side effect this routine returns to the bit loop just before
; the RRF instruction with carry set, which will roll a '1' into
; char_buf, which is then used as the stop bit when it rolls out
; again after being shifted 8 times.

	MOVLW .52 ; Initialise temp
	MOVWF temp

	MOVLW .1 ; Put 1 in W for incrementing temp

	GOTO $+1 ; Waste 2 cycles
	NOP ; Waste 1 cycle
delay_loop
	ADDWF temp ; Increment temp 1
	IFNC ; Did it overflow' 1
	GOTO delay_loop ; No: go round again 2
; ---
; time = 4 Tcy

	RETLW 0 ; Yes: return
; **************************************************************
get_put_char

; This subroutine deals with buffer operations, either storing a
; character from char_buf to the buffer or fetching it from the
; buffer. The routine uses buf_ptr (the logical buffer address)
; to calculate the physical address for the character.
;
; The 4 bit character will be stored at the current "memory
; location" in the buffer. The buffer is a large chunk of RAM
; from 0x10 to 0x1F and 0x30 to 0x33. Two "memory locations"
; are contained in each byte, one in each nibble.
;
; The variable buf_ptr points to the next free logical memory
; location, and the constants lo_mem and hi_mem record the
; physical start and end locations of the RAM block. The
; constant buf_sz holds the number of nibbles (or "memory
; locations") that can be filled.
;
; Note there is a discontinuity in the address space, which must
; be dealt with when mapping the logical memory location to the
; physical memory address
;
; To calculate the position in memory to store the current
; character, divide buf_ptr by 2 and add lo_mem to give the
; address in RAM. Next, check if this exceeds the 0x1F address
; range by checking bit 5 of the resultant address, and if
; necessary force the address into the 0x3n address space by
; setting bit 4. Check whether buf_ptr is odd or even by
; examining its LSB to see whether to store the character in the
; upper or lower nibble.

; When buf_ptr is zero it is pointing at the first "memory
; location", which is the low nibble of the first byte.
;
; All RAM in the memory buffer is cleared at the beginning of
; the main loop, so it is not necessary to clear each "memory
; location" before storing anything there.
;
; Note that for a 'put' operation the character arrives here in
; the upper nibble of char_buf, and for a 'get' operation the
; character is returned in the lower nibble.

	RRF buf_ptr,W ; load W with buf_ptr/2. Carry
; flag is rolled in.

	MOVWF FSR ; and use the FSR to point to it
; Upper bits of FSR are forced
; to 1 so junk in Carry flag
; doesn't matter.

	MOVLW lo_mem ; add the buffer start address
	ADDWF FSR ; to get the physical address to
; store the character

	IFSET FSR,RP0 ; Check for overflow into the
	BSF FSR,4 ; second register page and set
; bit 4 to move into 0x3n address
; space if necessary

	IFSET flag,read_buf ; check whether this is a read
	GOTO get_char 		; or write operation
put_char
	MOVF char_buf,W 	; Move the character (in high
						; nibble) to W
	IFCLR buf_ptr,LSB 	; except if LSB of buf_ptr is '0'
	SWAPF char_buf,W 	; then the destination is an even
						; nibble, so swap the character
; to the low nibble

	IORWF INDF 			; since the buffer was cleared
						; OR the character into place
	MOVLW buf_sz + 1 	; set limit for 'put' operation
						; to size of buffer
	GOTO get_put_done
get_char
	MOVF INDF,W 		; Fetch data from buffer to W
	IFSET buf_ptr,LSB 	; if LSB of buf_ptr is set the
	SWAPF INDF,W 		; desired character is an odd
						; nibble, so swap the nibbles
	ANDLW 0x0F 			; mask off upper nibble
	MOVWF char_buf 		; move it to the character buffer
	MOVF num_chr,W 		; set limit for 'get' operation
						; to number of characters read
						; from card
get_put_done
	INCF buf_ptr 		; increment memory pointer.
	XORWF buf_ptr,W 	; check if this was the last
						; nibble in the buffer by
						; comparing against W (either
						; buf_sz or num_chr)
	IFZ
	BSF flag,buf_end ; then set a flag
	RETLW 0
; **************************************************************
; Main Program starts here
; **************************************************************
start
	CLRF STATUS
	MOVLW B'11000000' 	; Disable GPIO pull-ups and wake
						; up feature
	OPTION
	MOVLW B'00001110' ; Set GPIO <1:3> as inputs...
	TRIS GPIO ; Note: GP3 is always input
	CLRF GPIO ; GPIO outputs all 0

if invert_tx
	BSF GPIO,ser_out ; except for invert_tx condition
endif
; Clear RAM from 0x07 to 0x0F
; for neatness

	MOVLW 0x07 ; Load start address (0x07) into
	MOVWF FSR ; the FSR
clrloop
	CLRF INDF 		; Clear the RAM location FSR is
					; pointing to
	INCF FSR ; Increment FSR to next location
	MOVF FSR,W ; Check if FSR is pointing past
	XORLW 0x10 | 0xC0 	; its end point. Remember MSBs
						; of FSR read '11'
	IFNZ ; If counter was not 0x10
	GOTO clrloop ; then loop again

; Main program loop is here
main_loop
; First clear memory buffer in
; the same way as above.
	MOVLW lo_mem ; Fetch buffer start address
	MOVWF FSR
clr_buf_loop
	CLRF INDF
	INCF FSR
	IFSET FSR,RP0 ; If FSR points to register page 1
	BSF FSR,4 			; set bit 4 to move into 0x3n
						; address space
	MOVF FSR,W ; Check for buffer end address.
	XORLW (hi_mem + 1) | 0xC0
	IFNZ ; If not end then loop around
	GOTO clr_buf_loop
	CLRF buf_ptr ; Initialise buffer pointer to 0
	MOVLW buf_sz ; Initialise the number of
	MOVWF num_chr 	; characters read to the maximum
					; in case of overflow later
	MOVLW start_code ; Initialise the LRC to the start
	MOVWF parityLRC ; sentinel code.
	CLRF flag ; Initialise control flags to
	BSF flag,bad_LRC 	; zero then set the bad_LRC
						; flag. Assume the LRC is bad
						; until the check at the end.
	MOVLW 'R' ; Send "Ready" from serial port
	CALL send_char
	MOVLW 'e'
	CALL send_char
	MOVLW 'a'
	CALL send_char
	MOVLW 'd'
	CALL send_char
	MOVLW 'y'
	CALL send_char
	MOVLW .13 ; Send CR LF from serial port
	CALL send_char
	MOVLW .10
	CALL send_char
	CLRF char_buf ; Clear character input buffer
wait_card
	IFSET GPIO,card ; Check ^CARD line
	GOTO wait_card ; if it's high then keep waiting
; ^CARD is low, so a card has started passing through the reader

wt_clk_lo
	IFSET GPIO,clock ; Check ^CLK line
	GOTO wt_clk_lo ; If it's high then keep waiting
; ^CLK is low, so valid data is on the ^DATA pin. If ^DATA is
; low the data bit on the card is a '1', so set carry and toggle
; the parity bit counter. If ^DATA is high the data bit on the
; card is a '0', so clear the carry. Roll the carry flag into
; the character buffer.
;
; All this data processing must be done as quickly as possible,
; but fortunately the card is being swiped by a human, so from
; the micro's point of view it is all happening very slowly.
chk_data
	IFSET GPIO,signal ; Check ^DATA
	GOTO data_0 ; If it's high, data bit is '0'
data_1
	BSF STATUS,C 	; Otherwise it's low so data bit
					; is '1', so set carry flag
	MOVLW 0x80 ; and toggle parity bit in
	XORWF parityLRC ; parityLRC register
	BTFSS STATUS,C 	; Use that fact that carry is
					; set to skip the next line.
data_0
	BCF STATUS,C ; bit is '0', so clear carry
store_bit
	RRF char_buf 	; shift data bit in carry flag
					; into the input buffer, and
					; shift LSB out into carry flag.
; If the start sentinel code has not yet been seen the LSB will
; have been '0', so carry will be '0'. If the start sentinel
; code has been seen then there will have been a sentinel bit
; set in char_buf which falls out after shifting 5 bits (one
; character) in.
	IFC ; So, check the carry flag
	GOTO got_char
; Otherwise, here a bit has just been read. Check if the start
; sentinel code has ever been seen. If it has then a sentinel
; bit will drop out after each character, which is dealt with by
; the code above. If not then it is necessary to check for the
; start sentinel code after reading each bit.
	IFSET flag,found_start; Has the start code been seen'
	GOTO wt_clk_hi 	; Yes, so wait for ^CLK to go
					; high again and get next bit.
; No, so check for the start
; sentinel code now

	MOVLW B'11111000' 	; The start code is five bits
						; long, so mask off the low 3
	ANDWF char_buf 		; bits in the buffer (which are
						; probably '0' anyway)
	MOVLW start_code ; and compare start_code
	XORWF char_buf,W ; to the buffer
	IFNZ ; Is it the start code'
	GOTO wt_clk_hi 		; No, so wait for ^CLK to go high
						; again and get the next bit.
	BSF flag,found_start	; Yes, so set a flag
							; Prepare for next character
next_char
	BCF parityLRC,MSB ; clear the parity flag,
	CLRF char_buf ; clear the input buffer,
	BSF char_buf,4 ; and set a sentinel bit
; Now wait for the next data bit
wt_clk_hi
	IFCLR GPIO,clock ; Check ^CLK line
	GOTO wt_clk_hi ; Keep waiting whilst it's low
	GOTO wt_clk_lo 	; Then go and wait for it to be
					; low again and get another bit
got_char
; Jump here when a sentinel bit has dropped out of the input
; buffer. There is now a character in the top five bits of
; char_buf, the low three bits will be '0', and carry will be '1'
; The MSB of char_buf is the parity bit for the character, which
; can now be discarded, and bits [6..3] form the character itself:
;
; char_buf 7 6 5 4 3 2 1 0
; P D3 D2 D1 D0 0 0 0
;
; First, check the parity of the character just read. The
; characters on the card are encoded with odd parity, and before
; each character is read the parity bit in parityLRC is cleared.
; This bit is toggled every time a '1' is read for the current
; character, which means that if the character was read correctly
; this bit will be '1'.
	IFCLR parityLRC,MSB ; If parity bit is '0'
	BSF flag,bad_parity ; set the parity error flag
; Now XOR char_buf with the parityLRC register to update the LRC
; state. The LRC portion of the parityLRC register was
; initialised with the code for the start sentinel. Every time
; a character is read it is XORed with the parityLRC register.
; Any bits in the char_buf register which are set will toggle
; bits in the LRC portion of parityLRC. If there have been no
; bit errors after reading all the characters and the LRC from
; the card, then the LRC bits in the parityLRC register will all
; be zero.
	MOVF char_buf,W ; Copy char_buf to W
	XORWF parityLRC ; XOR with the parityLRC register
					; to update the LRC calculation
; If the end sentinel has not
	IFCLR flag,found_end ; yet been seen then this is
	GOTO not_LRC ; not the LRC, so store it
	MOVF parityLRC,W ; Otherwise it was the LRC, so
	ANDLW b'01111000' 	; get the LRC check from the
						; parityLRC register, and mask
						; off the parity flag
	IFZ ; If it is zero then the LRC was
	BCF flag,bad_LRC ; okay so clear the bad_LRC flag
	MOVF buf_ptr,W ; Copy the value of buffer pointer
	MOVWF num_chr ; to num_chr
	GOTO dump_buffer ; and dump it out
not_LRC
	MOVLW end_code ; Is this the end sentinel'
	XORWF char_buf,W
	IFZ ; If so, the next character is
	BSF flag,found_end ; the LRC, so set a flag
	IFZ
	GOTO next_char ; and don't bother storing it
	RLF char_buf 	; discard parity by shifting it
					; out, leaving the 4 bit
					; character in the upper nibble
	MOVLW 0xF0 ; mask off the lower nibble
	ANDWF char_buf
	CALL get_put_char ; and store the character
	IFCLR flag,buf_end ; Is the buffer full'
	GOTO next_char ; no, so get the next character
; Otherwise, fall through...

; Jump to dump_buffer when the buffer is full. This routine
; loops through each "location" in the memory buffer, and sends
; the character at that location out serially on an output pin.
dump_buffer
	CLRF buf_ptr ; Load buffer pointer with 0
	BSF flag,read_buf ; Set the flag to read mode
	BCF flag,buf_end ; Clear the buf_end flag
loop_buffer
	CALL get_put_char ; Get character from buffer
	MOVLW .48 ; convert to ASCII by adding 48
	ADDWF char_buf, W ; and put the result in W
	CALL send_char ; and send the character
	IFCLR flag,buf_end ; have we emptied the buffer'
	GOTO loop_buffer ; No, so loop around
; After sending the contents of the buffer, two more characters
; are sent to indicate any errors. If there was bad parity on
; any character a "P" is sent, and if there was a bad LRC an "L"
; is sent. If either condition was okay we send a period "."
	MOVLW '.' ; Load ASCII "." into W
	IFSET flag,bad_parity ; If parity was ever bad
	MOVLW 'P' ; load ASCII "P" into W instead
	CALL send_char ; then send the character
	MOVLW '.' ; Load ASCII "." into W again
	IFSET flag,bad_LRC ; If LRC was bad
	MOVLW 'L' ; load ASCII 'L' into W instead
	CALL send_char ; and send the character
	MOVLW .13 ; Send CR LF from serial port
	CALL send_char
	MOVLW .10
	CALL send_char
	GOTO main_loop 	; Back to the beginning and wait
					; for another card.

; The end is nigh...
END
